"""
--- Day 14: Regolith Reservoir ---
https://adventofcode.com/2022/day/14
"""
import os
from aocd import data
from aoc_wim.zgrid import ZGrid
from aoc_wim.zgrid import zline
from aoc_wim.search import BFS

grid = ZGrid()
for line in data.splitlines():
    zs = [complex(*map(int, xy.split(","))) for xy in line.split(" -> ")]
    for z0, z1 in zip(zs, zs[1:]):
        for z in zline(z0, z1):
            grid[z] = "#"


def s1(s0):
    # how will piece of sand at s0 come to rest (if anywhere)
    path = [s0]
    while path[-1].imag <= ymax:
        for dx in 0, -1, 1:
            if path[-1] + 1j + dx not in grid:
                path.append(path[-1] + 1j + dx)
                break
        else:
            break
    return path


def flow(render=os.environ.get("AOC_DEBUG")):
    path = []
    while True:
        path = s1(path[-1] if path else z_source)
        s = path.pop()
        if s.imag > ymax or s in grid:  # flowing into void, or flow is blocked
            return
        grid[s] = "o"
        if render:
            grid.draw(overlay={z_source: "+", s: "O"}, clear=True, title=__doc__)


z_source = 500
ymax = int(grid.bottom_right.imag)
flow()
a = grid.count("o")


def adj(z0):
    for dx in 0, -1, 1:
        z = z0 + dx + 1j
        if grid.get(z) != "#" and z.imag < ymax + 2:
            yield z


bfs = BFS(adj)
bfs(z_source)
b = len(bfs.seen)

print("part a:", a)
print("part b:", b)



"""
--- Day 23: Opening the Turing Lock ---
https://adventofcode.com/2015/day/23
"""
from aocd import data


class Computer:
    def __init__(self, instructions=""):
        self.reset(instructions=instructions)

    def reset(self, i=0, instructions=""):
        self.i = i
        self.instructions = instructions.strip().splitlines()
        self.registers = {}.fromkeys("ab", 0)

    def compute(self, i=0, verbose=False):
        self.i = i
        while 0 <= self.i < len(self.instructions):
            line = self.instructions[self.i]
            op, args = line.split(None, maxsplit=1)
            args = args.split(", ")
            args = [x if x in self.registers else int(x) for x in args]
            f = getattr(self, op)
            if verbose:
                registers = ["{}:{}".format(k, v) for k, v in self.registers.items()]
                state = ",".join(registers)
                print(state, f"L{i}", op, *args)
            f(*args)
            self.i += 1

    def hlf(self, r):
        self.registers[r] //= 2

    def tpl(self, r):
        self.registers[r] *= 3

    def inc(self, r):
        self.registers[r] += 1

    def jmp(self, offset):
        self.i += offset - 1

    def jie(self, r, offset):
        if self.registers[r] % 2 == 0:
            self.jmp(offset)

    def jio(self, r, offset):
        if self.registers[r] == 1:
            self.jmp(offset)


if __name__ == "__main__":
    computer = Computer()

    computer.reset(instructions=data)
    computer.compute()
    print(computer.registers["b"])

    computer.reset(instructions=data)
    computer.registers["a"] = 1
    computer.compute()
    print(computer.registers["b"])



"""
--- Day 11: Chronal Charge ---
https://adventofcode.com/2018/day/11
"""
import numpy as np
from aocd import data
from scipy.signal import convolve2d


def gen_grid(data):
    d = int(data or 0)
    n = 300
    a = np.empty((n, n), dtype=int)
    v = np.arange(0, n, dtype=int) + 1
    a[:] = v + 10
    a[:] *= v.reshape(-1, 1)
    a += d
    a[:] *= v + 10
    a %= 1000
    a //= 100
    a -= 5
    return a


def max_power(grid, kernels=(3,)):
    maxs = []
    for k in kernels:
        kernel = np.ones((k, k), dtype=int)
        c = convolve2d(grid, kernel, mode="valid")
        y, x = np.unravel_index(c.argmax(), c.shape)
        maxs.append((c[y, x], x + 1, y + 1, k))
    return max(maxs)


grid = gen_grid(data)

_, x, y, _ = max_power(grid)
print("part a:", f"{x},{y}")

_, x, y, k = max_power(grid, kernels=range(1, 20))
print("part b:", f"{x},{y},{k}")



"""
--- Day 17: No Such Thing as Too Much ---
https://adventofcode.com/2015/day/17
"""
from collections import Counter

from aocd import data

from aoc_wim.stuff import subset_sum


def part_a(vals, target, impl=subset_sum):
    return sum(1 for subset in impl(vals, target))


def part_b(vals, target, impl=subset_sum):
    counter = Counter(len(subset) for subset in impl(vals, target))
    return counter[min(counter)]


if __name__ == "__main__":
    vals = [int(n) for n in data.splitlines()]
    print(part_a(vals, target=150))
    print(part_b(vals, target=150))



"""
--- Day 24: It Hangs in the Balance ---
https://adventofcode.com/2015/day/24
"""
from collections import Counter
from itertools import combinations
from math import prod

from aocd import data

from aoc_wim.stuff import subset_sum


def parsed(data, n_groups):
    vals = [int(n) for n in data.strip().split()]
    total = sum(vals)
    if total % n_groups != 0:
        raise Exception
    return vals, total // n_groups


def bag_sub(list_big, sublist):
    counter = Counter(sublist)
    result = []
    for k in list_big:
        counter[k] -= 1
        if counter[k] < 0:
            result.append(k)
    return result


def partitions(vals, target):
    for group in subset_sum(vals, target):
        rem = bag_sub(vals, group)
        groups = partitions(rem, target) if sum(rem) > target else (rem,)
        yield from ((group, group_) for group_ in groups)


def get_shortest_sums(vals, target):
    results = []
    for i in range(len(vals)):
        for comb in combinations(vals, i):
            if sum(comb) == target:
                results.append(comb)
        if results:
            return results
        del results[:]


def solve(data, n_groups):
    vals, target = parsed(data, n_groups=n_groups)
    group1s = get_shortest_sums(vals, target)
    group1s.sort(key=prod)

    for group1 in group1s:
        rem = bag_sub(vals, group1)
        gen = partitions(rem, target)
        try:
            next(gen)
        except StopIteration:
            continue
        return prod(group1)


print(solve(data, n_groups=3))
print(solve(data, n_groups=4))



"""
--- Day 18: Snailfish ---
https://adventofcode.com/2021/day/18
"""
from aocd import data
import json


class SnailfishNumber:
    """shh, it's really a binary tree"""

    def __init__(self, val, parent=None):
        self.parent = parent
        if isinstance(val, str):
            val = json.loads(val)
        if isinstance(val, int):
            self.left = self.right = None
            self._magnitude = val
        else:
            left, right = val
            if not isinstance(left, SnailfishNumber):
                left = SnailfishNumber(left, parent=self)
            if not isinstance(right, SnailfishNumber):
                right = SnailfishNumber(right, parent=self)
            self.left = left
            self.right = right

    def __abs__(self):
        if self.left is self.right is None:
            return self._magnitude
        return 3 * abs(self.left) + 2 * abs(self.right)

    def __add__(self, other):
        if not isinstance(other, SnailfishNumber):
            return NotImplemented
        result = SnailfishNumber(f"[{self},{other}]")
        self.parent = other.parent = result
        result.reduce()
        return result

    def __radd__(self, other):
        if other == 0:
            # this is to allow summing a homogeneous list of SnailfishNumbers
            return self
        return NotImplemented

    def __str__(self):
        if self.leaf:  # regular number
            return str(abs(self))
        return f"[{self.left},{self.right}]"

    def __eq__(self, other):
        if not isinstance(other, SnailfishNumber):
            return NotImplemented
        return str(self) == str(other)

    @property
    def leaf(self):
        return self.left is self.right is None

    def reduce(self):
        # "explode": prune nodes at depth >= 4
        # "split": grow new nodes at values >= 10
        while True:
            stack = [(self, 0)]
            leaf_nodes = []
            explode = split = None
            while stack:
                node, depth = stack.pop()
                if node.leaf:
                    leaf_nodes.append(node)
                    if abs(node) >= 10:
                        split = node
                else:
                    stack.extend([(node.left, depth + 1), (node.right, depth + 1)])
                    if depth == 4 and node.left.leaf and node.right.leaf:
                        explode = node
            if explode:
                node = explode
                assert not node.leaf
                # Exploding pairs will always consist of two regular numbers.
                assert node.left.leaf and node.right.leaf
                # distribute pair values into prev and next leaf nodes, if any
                for i, n in enumerate(leaf_nodes):
                    if node.right is n:
                        if i > 0:
                            next_leaf = leaf_nodes[i - 1]
                            next_leaf._magnitude += abs(node.right)
                    if node.left is n:
                        if i + 1 < len(leaf_nodes):
                            prev_leaf = leaf_nodes[i + 1]
                            prev_leaf._magnitude += abs(node.left)
                # convert current node into a leaf
                node._magnitude = 0
                node.left = node.right = None
                # print("after explode: ", self)
            elif split:
                node = split
                assert node.leaf
                n = abs(node)
                assert n >= 10
                node.left = SnailfishNumber(n // 2, parent=self)
                node.right = SnailfishNumber((n + 1) // 2, parent=self)
                # print("after split:   ", self)
            else:  # reduction is finished
                break


ns = [SnailfishNumber(line) for line in data.splitlines()]
print("part a:", abs(sum(ns)))
print("part b:", max([abs(n1 + n2) for n1 in ns for n2 in ns], default=0))



"""
--- Day 18: Operation Order ---
https://adventofcode.com/2020/day/18
"""
from aocd import data
from collections import deque


def evaluate(d, part="a"):
    op = "*"
    val = None
    prev = 1
    while d:
        tok = d.popleft()
        if tok == "(":
            val = evaluate(d, part=part)
        elif tok.isdigit():
            val = int(tok)
        elif tok in "+*":
            op = tok
        elif tok == ")":
            return prev
        if val is not None:
            if op == "+":
                prev += val
            elif op == "*":
                if part == "b":
                    while d and d[0] == "+":
                        d.popleft()
                        r = d.popleft()
                        if r.isdigit():
                            val += int(r)
                        elif r == "(":
                            val += evaluate(d, part=part)
                prev *= val
            val = None
    return prev


lines = data.replace(" ", "").splitlines()
print("part a:", sum([evaluate(deque(line), part="a") for line in lines]))
print("part b:", sum([evaluate(deque(line), part="b") for line in lines]))



"""
--- Day 20: Jurassic Jigsaw ---
https://adventofcode.com/2020/day/20
"""
import math
from aocd import data
from aoc_wim.zgrid import ZGrid
import networkx as nx
import numpy as np
from parse import parse
from itertools import combinations


class Tile:
    opposite = {
        "L": "R",
        "R": "L",
        "U": "D",
        "D": "U",
    }

    def __init__(self, id, data):
        self.id = id
        self.data = data
        self.angle = 0
        self.flip = False

    @property
    def U(self):
        return *self.data[0],

    @property
    def D(self):
        return *self.data[-1],

    @property
    def L(self):
        return *self.data[:,0],

    @property
    def R(self):
        return *self.data[:,-1],

    @property
    def edges(self):
        return {self.U, self.D, self.L, self.R}

    def fits_with(self, other):
        return any({e, e[::-1]} & self.edges for e in other.edges)

    def transform(self):
        self.data = np.rot90(self.data)
        self.angle += 90
        if self.angle == 360:
            self.angle = 0
            self.data = np.flipud(self.data)
            self.flip = not self.flip

    def orient_to(self, other, side="R"):
        # line up the left edge of this tile with the right edge of other
        # return False if tiles can not be oriented
        edge = getattr(other, side)
        opposite_side = Tile.opposite[side]
        for i in range(8):
            if edge == getattr(self, opposite_side):
                return True
            self.transform()
        return False

    def trimmed(self):
        return self.data[1:-1, 1:-1]


tiles_raw = data.split("\n\n")
tiles = {}
for tile in tiles_raw:
    pre, tile = tile.split("\n", 1)
    tile_id, = parse("Tile {:d}:", pre).fixed
    A = np.array([[*r] for r in tile.splitlines()])
    A = (A == "#").astype(int)
    tiles[tile_id] = Tile(tile_id, A)

g = nx.Graph()
for t0, t1 in combinations(tiles.values(), 2):
    if t0.fits_with(t1):
        g.add_edge(t0, t1)

corners = {t.id for t in tiles.values() if len(list(g.neighbors(t))) == 2}
a = math.prod(corners)
print("part a:", a)

# start joining from any old corner
corner_id = corners.pop()
tile00 = tiles.pop(corner_id)
# corner tiles should have two neighbors
x, y = g.neighbors(tile00)
# orient first corner so that it fits top-left
# keep poking the tile until the top and left edges can't match with neighbors
while True:
    top = (*tile00.data[0],)
    left = (*tile00.data[:,0],)
    both = {top, left, top[::-1], left[::-1]}
    if not both & (x.edges | y.edges):
        break
    tile00.transform()

n = math.isqrt(len(g))  # we have n*n square of tiles
grid = {(0, 0): tile00}
for row in range(n):
    for col in range(n):
        if row == col == 0:
            continue
        try:
            prev = grid[row, col-1]  # tile on left
            side = "R"
        except KeyError:
            prev = grid[row-1, col]  # tile above
            side = "D"
        for tile in {*g.neighbors(prev)}.intersection(tiles.values()):
            if tile.orient_to(prev, side):
                grid[row, col] = tile
                del tiles[tile.id]
                break

full_rows = [np.hstack([grid[r, c].trimmed() for c in range(n)]) for r in range(n)]
full_grid = np.vstack(full_rows)
big_tile = Tile(0, full_grid)

sea_monster_raw = """\
                  #
#    ##    ##    ###
 #  #  #  #  #  #
"""
sea_monster = (np.array(ZGrid(sea_monster_raw)) == "#").astype(int)
n = sea_monster.sum()
h, w = sea_monster.shape
H, W = big_tile.data.shape

n_monsters = 0
while n_monsters <= 0:
    big_tile.transform()
    n_monsters = 0
    for row in range(H - h):
        for col in range(W - w):
            section = big_tile.data[row:row + h, col:col + w]
            if (sea_monster * section).sum() == n:
                n_monsters += 1

print("part b:", big_tile.data.sum() - n_monsters * sea_monster.sum())



"""
--- Day 19: Not Enough Minerals ---
https://adventofcode.com/2022/day/19
"""
from aocd import data
from parse import parse
from typing import NamedTuple


template = (
    "Blueprint {id:d}: "
    "Each ore robot costs {ore_ore:d} ore. "
    "Each clay robot costs {cla_ore:d} ore. "
    "Each obsidian robot costs {obs_ore:d} ore and {obs_cla:d} clay. "
    "Each geode robot costs {geo_ore:d} ore and {geo_obs:d} obsidian."
)
if len(data.split("\n\n")) == 2:
    lines = data.replace("  ", " ").replace("\n\n", " --- ").replace("\n", "").split(" --- ")
else:
    lines = data.splitlines()


class Node(NamedTuple):
    n_geo: int = 0
    n_obs: int = 0
    n_cla: int = 0
    n_ore: int = 0

    r_geo: int = 0
    r_obs: int = 0
    r_cla: int = 0
    r_ore: int = 1

    def buy(self, rtype=None):
        n_geo = self.n_geo + self.r_geo
        n_obs = self.n_obs + self.r_obs
        n_cla = self.n_cla + self.r_cla
        n_ore = self.n_ore + self.r_ore
        r_geo = self.r_geo
        r_obs = self.r_obs
        r_cla = self.r_cla
        r_ore = self.r_ore
        if rtype == "geo":
            r_geo += 1
        elif rtype == "obs":
            r_obs += 1
        elif rtype == "cla":
            r_cla += 1
        elif rtype == "ore":
            r_ore += 1
        if rtype is not None:
            cost_obs, cost_cla, cost_ore = bp[f"cost_{rtype}"]
            n_obs -= cost_obs
            n_cla -= cost_cla
            n_ore -= cost_ore
        return Node(n_geo, n_obs, n_cla, n_ore, r_geo, r_obs, r_cla, r_ore)


blueprints = []
rtypes = "ore", "cla", "obs", "geo"
for line in lines:
    parsed = parse(template, line).named
    bp = {"pk": parsed["id"]}
    costs = []
    for r0 in rtypes:
        k = f"cost_{r0}"
        cost = parsed.get(f"{r0}_obs", 0), parsed.get(f"{r0}_cla", 0), parsed.get(f"{r0}_ore", 0)
        bp[k] = cost
        costs.append(cost)
    bp["max_ore"] = max(c[2] for c in costs)
    bp["max_cla"] = max(c[1] for c in costs)
    bp["max_obs"] = max(c[0] for c in costs)
    blueprints.append(bp)


def adj(n0):
    result = []
    c_obs, c_cla, c_ore = bp["cost_geo"]
    if c_obs <= n0.n_obs and c_ore <= n0.n_ore:
        result.append(n0.buy(rtype="geo"))

    c_obs, c_cla, c_ore = bp["cost_obs"]
    if c_cla <= n0.n_cla and c_ore <= n0.n_ore and n0.r_geo <= 2 and n0.r_obs < bp["max_obs"]:
        result.append(n0.buy(rtype="obs"))

    c_obs, c_cla, c_ore = bp["cost_cla"]
    if c_ore <= n0.n_ore and n0.r_obs <= 2 and n0.r_geo == 0 and n0.r_cla < bp["max_cla"]:
        result.append(n0.buy(rtype="cla"))

    c_obs, c_cla, c_ore = bp["cost_ore"]
    if c_ore <= n0.n_ore and n0.r_cla <= 2 and n0.r_obs == 0 and n0.r_geo == 0 and n0.r_ore < bp["max_ore"]:
        result.append(n0.buy(rtype="ore"))

    result.append(n0.buy())  # buy nothing - increases budget and elapsed time
    return result


class DFS:

    def __init__(self, adj, max_depth=None):
        self.adj = adj
        self.seen = {}
        self.max_depth = max_depth

    def __call__(self, s0, target=None):
        stack = [(0, s0)]
        maxd = self.max_depth or float("inf")
        seen = self.seen
        best = 0
        while stack:
            depth, state = stack.pop()
            if state in seen and seen[state] <= depth:
                # already got here by a more efficient path - don't bother extending
                continue
            seen[state] = depth
            if state.n_geo > best:
                best = state.n_geo
                print(f"new best {best} - {len(seen)} {len(stack)} depth={depth}, {state}")
            if depth + 1 <= maxd:
                t_remain = maxd - depth
                potential = state.n_geo + t_remain * state.r_geo + (t_remain * (t_remain - 1))//2
                if potential >= best:
                    stack += [(depth + 1, s) for s in self.adj(state)[::-1]]
        return best


a = 0
for bp in blueprints:
    bfs = DFS(adj, max_depth=24)
    bfs(Node())
    best = max([s.n_geo for s, d in bfs.seen.items() if d == 24])
    print(f"n_geodes={best} blueprint={bp}")
    a += bp["pk"] * best
print("part a:", a)

b = 1
for bp in blueprints[:3]:
    bfs = DFS(adj, max_depth=32)
    bfs(Node())
    best = max([s.n_geo for s, d in bfs.seen.items() if d == 32])
    print(f"n_geodes={best} blueprint={bp}")
    b *= best
print("part b:", b)



"""
--- Day 19: Medicine for Rudolph ---
https://adventofcode.com/2015/day/19
"""
from collections import defaultdict

from aocd import data


reactions, element = data.split("\n\n")
tr = defaultdict(list)
tri = {}
for line in reactions.splitlines():
    s, r = line.split(" => ")
    tr[s].append(r)
    tri[r] = s


seen = set()
for s, rs in tr.items():
    splitted = element.split(s)
    for i, (left, right) in enumerate(zip(splitted, splitted[1:])):
        for r in rs:
            new = splitted[:]
            new[i : i + 2] = [left + r + right]
            new = s.join(new)
            seen.add(new)


replacements = 0
while element != "e":
    pos = {}
    for k, v in tri.items():
        delta = len(k) - len(v)
        if k in element:
            if v != "e" or len(element) - delta == 1:
                pos[k] = (element.rfind(k) + len(k), delta)
    k = max(pos, key=pos.get)
    v = tri[k]
    # replace from right
    element = element[::-1].replace(k[::-1], v[::-1], 1)[::-1]
    replacements += 1


print("part a:", len(seen))
print("part b:", replacements)



"""
--- Day 2: Corruption Checksum ---
https://adventofcode.com/2017/day/2
"""
from itertools import combinations

from aocd import data


checksum = result = 0
for line in data.splitlines():
    row = [int(x) for x in line.split()]
    checksum += max(row) - min(row)
    for x, y in combinations(row, 2):
        denominator, numerator = sorted([x, y])
        quotient, remainder = divmod(numerator, denominator)
        if not remainder:
            result += quotient
            break

print("part a:", checksum)
print("part b:", result)



"""
--- Day 4: High-Entropy Passphrases ---
https://adventofcode.com/2017/day/4
"""
from aocd import data


def has_dupe(txt):
    parts = txt.split()
    return len(parts) != len(set(parts))


def has_anagram(txt):
    parts = [tuple(sorted(x)) for x in txt.split()]
    return len(parts) != len(set(parts))


a = b = 0
for txt in data.splitlines():
    a += not has_dupe(txt)
    b += not has_anagram(txt)

print("part a:", a)
print("part b:", b)



"""
--- Day 23: Coprocessor Conflagration ---
https://adventofcode.com/2017/day/23
"""
from aocd import data
from aoc_wim.aoc2017 import Comp

# import logging; logging.basicConfig(format="%(message)s", level=logging.DEBUG)

comp_a = Comp(data)
while not comp_a.blocked():
    comp_a.step()
print("part a:", comp_a.n_mul)

comp_b = Comp(data, a0=1)
comp_b.code[8:10] = ("wtf", "f", "b"), ("jnz", "1", "15")
while not comp_b.blocked():
    comp_b.step()
print("part b:", comp_b.reg["h"])



"""
--- Day 17: Conway Cubes ---
https://adventofcode.com/2020/day/17
"""
from aocd import data
import numpy as np
from scipy.signal import convolve


def evolve(A, n=6):
    kernel = np.ones((3,) * A.ndim, dtype=A.dtype)
    kernel[(1,) * A.ndim] = 0  # hollow center
    for _ in range(n):
        C = convolve(A, kernel)
        A = np.pad(A, pad_width=1)
        A = ((A == 1) & ((C == 2) | (C == 3))) | ((A == 0) & (C == 3))
        A = A.astype(int)
    return A


A0 = (np.array([[*r] for r in data.splitlines()]) == "#").astype(int)
print("part a:", evolve(A0[..., None]).sum())
print("part b:", evolve(A0[..., None, None]).sum())



"""
--- Day 4: Security Through Obscurity ---
https://adventofcode.com/2016/day/4
"""
from collections import Counter
from collections import deque
from string import ascii_lowercase

from aocd import data
from parse import parse


def decrypt_name(encrypted_name, sector_id):
    d = deque(ascii_lowercase)
    d.rotate(-sector_id)
    lookup = {"-": " "}
    lookup.update(zip(ascii_lowercase, d))
    decrypted_name = "".join(lookup[k] for k in encrypted_name)
    return decrypted_name


sector_id_sum = 0
northpole_room = None
template = "{}-{:d}[{}]"
for line in data.splitlines():
    encrypted_name, sector_id, checksum = parse(template, line)
    counter = Counter(encrypted_name.replace("-", ""))
    most_common = sorted(counter, key=lambda k: (-counter[k], k))
    check = "".join(most_common[:5])
    if check == checksum:
        sector_id_sum += sector_id
    decrypted_name = decrypt_name(encrypted_name, sector_id)
    if "pole" in decrypted_name:
        northpole_room = sector_id


print("part a:", sector_id_sum)
print("part b:", northpole_room)



"""
--- Day 16: Proboscidea Volcanium ---
https://adventofcode.com/2022/day/16
"""
from aocd import data
from collections import deque
import networkx as nx
import itertools
import heapq


valves = {}
graph = nx.Graph()
for line in data.splitlines():
    parts = line.split()
    f = int(parts[4].split("=")[1][:-1])
    v, *vs = [x.rstrip(",") for x in parts if x.rstrip(",").isupper()]
    if f > 0:
        valves[v] = f
    for t in vs:
        graph.add_edge(v, t)

dist = {}
for p1, p2 in itertools.combinations(["AA", *valves], 2):
    dist[p1, p2] = dist[p2, p1] = nx.shortest_path_length(graph, p1, p2)


def go(dont_visit=(), max_visits=float("inf")):
    # state vector:
    # current minute (int), current position (str), valves opened: minute (dict)
    s0 = 0, "AA", {}
    result = []
    q = deque([s0])
    while q:
        m, pos, v = q.popleft()
        if m >= M:
            result.append(v)
            continue
        candidates = [p for p in valves if p not in v and p not in dont_visit]
        if not candidates or len(v) >= max_visits:
            result.append(v)
            continue
        for p in candidates:
            d = dist[pos, p]
            if m + d + 1 > M:
                q.append((M, p, v))
            else:
                q.append((m + d + 1, p, {**v, p: m + d + 1}))
    return result


def score(end):
    return sum((M - v) * valves[k] for k, v in end.items())


M = 30
paths = go()
print("part a:", max(score(p) for p in paths))

M = 26
b = 0
your_paths = go(max_visits=len(valves) // 2)
for your_path in heapq.nlargest(20, your_paths, key=score):
    your_score = score(your_path)
    ele_paths = go(dont_visit=your_path)
    for ele_path in ele_paths:
        ele_score = score(ele_path)
        b = max(b, your_score + ele_score)
print("part b:", b)



"""
--- Day 13: Distress Signal ---
https://adventofcode.com/2022/day/13
"""
from aocd import data
from json import loads


class Packet:

    def __init__(self, content):
        self.content = content

    def __getitem__(self, index):
        return Packet(self.content[index])

    def __lt__(self, other):
        if not isinstance(other, Packet):
            return NotImplemented
        match self.content, other.content:
            case int(), int():
                if self.content < other.content:
                    return True
                elif other.content < self.content:
                    return False
                # raise NotImplementedError
                return False
            case list(), list():
                for x, y in zip(self, other):
                    if x < y:
                        return True
                    elif y < x:
                        return False
                if len(self.content) < len(other.content):
                    return True
                elif len(other.content) < len(self.content):
                    return False
                # raise NotImplementedError
                return False
            case list(), int():
                return self < Packet([other.content])
            case int(), list():
                return Packet([self.content]) < other

    def __repr__(self):
        return f"Packet({self.content})"


a = 0
i2 = 1
i6 = 2
for i, chunk in enumerate(data.split("\n\n"), 1):
    first, second = [Packet(loads(x)) for x in chunk.split("\n")]
    a += i * (first < second)
    for packet in first, second:
        if packet < Packet([[2]]):
            i2 += 1
            i6 += 1
        elif packet < Packet([[6]]):
            i6 += 1

print("part a:", a)
print("part b:", i2 * i6)



"""
--- Day 11: Space Police ---
https://adventofcode.com/2019/day/11
"""
from aocd import data
from aoc_wim.aoc2019 import IntComputer
from aoc_wim.ocr import AOCR
from collections import deque


class Robot:
    def __init__(self, data, colour0=0):
        self.brain = IntComputer(data, inputs=[colour0])
        self.brain.output = deque(maxlen=2)
        self.position = 0
        self.direction = -1j
        self.painted = {}

    def paint(self):
        while True:
            self.brain.run(until=IntComputer.op_output)
            self.brain.run(until=IntComputer.op_output)
            turn, colour = self.brain.output
            self.painted[self.position] = colour
            self.direction *= [-1j, 1j][turn]
            self.position += self.direction
            self.brain.input.append(self.painted.get(self.position, 0))

    def paint_until_halt(self):
        try:
            self.paint()
        except IntComputer.Halt:
            return


robot = Robot(data)
robot.paint_until_halt()
print("part a:", len(robot.painted))

robot = Robot(data, 1)
robot.paint_until_halt()
rego = AOCR[robot.painted]
print("part b:", rego)



import argparse
import ast
import shutil
import sys
import tempfile
import webbrowser
from pathlib import Path
from datetime import datetime
from textwrap import dedent

from aocd import get_data
from aocd.models import Puzzle
from aocd.utils import AOC_TZ
from aocd.utils import blocker


here = Path(__file__).parent.resolve()


def get_module_docstring(path):
    """get a .py file docstring, without actually executing the file"""
    with open(path) as f:
        return ast.get_docstring(ast.parse(f.read()))


def set_module_docstring(path, text):
    """update docstring of file at path with text, cautiously"""
    prev = get_module_docstring(path)
    if prev == text:
        # nothing to do
        return
    if prev is not None:
        print(repr(text))
        print(repr(prev))
        raise Exception(f"refusing to clobber existing docstring on {path}")
    content = Path(path).read_text()
    Path(path).write_text(f'"""\n{text}\n"""\n{content}')


def set_docstrings(files=()):
    """puts name and link into puzzle source files"""
    for file in files or here.glob("aoc*/q*.py"):
        day = int(file.name[1:3])
        year = int(file.parent.name[3:])
        puzzle = Puzzle(year, day)
        docstring = f"--- Day {day}: {puzzle.title} ---\n"
        docstring += puzzle.url
        set_module_docstring(file, docstring)


def start():
    """init a new source file at ./aocYYYY/qDD.py"""
    aoc_now = datetime.now(tz=AOC_TZ)
    years = range(2015, aoc_now.year + int(aoc_now.month >= 11))
    days = range(1, 26)
    parser = argparse.ArgumentParser(description="init current day")
    parser.add_argument(
        "day",
        nargs="?",
        type=int,
        default=min(aoc_now.day, 25) if aoc_now.month == 12 else 1,
        help="1-25 (default: %(default)s)",
    )
    parser.add_argument(
        "year",
        nargs="?",
        type=int,
        default=years[-1],
        help="2015-%(default)s (default: %(default)s)",
    )
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="create the template file even if source already exists",
    )
    args = parser.parse_args()
    if args.day in years and args.year in days:
        # be forgiving
        args.day, args.year = args.year, args.day
    if args.day not in days or args.year not in years:
        parser.print_usage()
        parser.exit(1)
    year = args.year
    day = args.day
    here = Path(__file__).parent
    src = here / f"aoc{year}" / f"q{day:02d}.py"
    block = False
    if day < 25 and year == aoc_now.year and day == aoc_now.day:
        if src.exists() and str(day) not in sys.argv:
            day += 1
            block = True
            src = here / f"aoc{year}" / f"q{day:02d}.py"
    if src.exists():
        if not args.force:
            sys.exit(f"{src} already exists! (use -f to overwrite)")
        shutil.copy2(src, tempfile.gettempdir())
    if day == 1 and not src.parent.is_dir():
        src.parent.mkdir()
        src.parent.joinpath("__init__.py").touch()
    src.write_text(dedent('''\
        from aocd import data
        from aoc_wim.autoparse import parsed
        from collections import Counter, defaultdict, deque
        d = parsed(data)
        if d != data:
            print(d)
            print(f"{parsed.n_bytes} bytes/{parsed.n_lines} lines, parsed by {parsed.parser}")

        # import numpy as np
        # import networkx as nx
        # from aoc_wim.zgrid import ZGrid
        # from aoc_wim import stuff

        # import logging; logging.basicConfig(level=logging.DEBUG)






        print("part a:", )
        print("part b:", )
    '''))
    test = here.parent / "tests" / str(year) / str(day).zfill(2) / "a.txt"
    if not test.exists():
        test.parent.mkdir(parents=True, exist_ok=True)
        test.write_text("\n-\n-\n")
    if block:
        blocker()
    data = get_data(day=day, year=year, block=True)
    print(data)
    set_docstrings([src])
    puzzle = Puzzle(year, day)
    if test.read_text() == "\n-\n-\n":
        test.write_text("{}\n-\n-\n".format(puzzle.example_data))
    webbrowser.open(puzzle.url)



"""
--- Day 10: Balance Bots ---
https://adventofcode.com/2016/day/10
"""
from math import prod
from aocd import data
from parse import parse


class Bot(object):
    def __init__(self, name, low_to=None, high_to=None, lh=(17, 61)):
        self.name = name
        self.low_to = low_to
        self.high_to = high_to
        self.chips = []
        self.part_b = False
        self.lh = lh

    def receive(self, chip):
        self.chips.append(chip)
        if len(self.chips) == 2:
            self.send()

    def send(self):
        low, high = sorted(self.chips)
        self.chips = []
        self.low_to.receive(low)
        self.high_to.receive(high)
        if (low, high) == self.lh:
            self.part_b = True


def parsed(data, lh=(17, 61)):
    receives = []
    receivers = {}
    for line in data.splitlines():
        if line.startswith("value"):
            tup = parse("value {:d} goes to {}", line).fixed
            receives.append(tup)
        else:
            result = parse("{} gives low to {} and high to {}", line)
            bot_name, low_to, high_to = result.fixed
            bot = receivers.get(bot_name, Bot(bot_name, lh=lh))
            bot.low_to = receivers.get(low_to, Bot(low_to))
            bot.high_to = receivers.get(high_to, Bot(high_to))
            for receiver in bot, bot.low_to, bot.high_to:
                receivers[receiver.name] = receiver
    return receives, receivers


def part_ab(data, lh=(17, 61)):
    receives, receivers = parsed(data, lh=lh)
    for n, bot_name in receives:
        receivers[bot_name].receive(n)
    outputs = [receivers[f"output {o}"].chips for o in (0, 1, 2)]
    [bot_a] = [v for v in receivers.values() if v.part_b]
    a = int(bot_a.name.lstrip("bot "))
    b = prod([n for [n] in outputs])
    return a, b


if __name__ == "__main__":
    a, b = part_ab(data)
    print("part a:", a)
    print("part b:", b)



"""
--- Day 10: Knot Hash ---
https://adventofcode.com/2017/day/10
"""
from functools import reduce
from operator import xor

from aocd import data


def munge(state, lengths, n=256, iterations=64):
    skip_size = 0
    current_position = 0
    for i in range(iterations):
        for length in lengths:
            for i in range(length // 2):
                i1 = (current_position + i) % n
                i2 = (current_position + length - i - 1) % n
                state[i1], state[i2] = state[i2], state[i1]
            current_position += length + skip_size
            skip_size += 1


def part_a(data, n=256):
    lengths = [int(x) for x in data.split(",")]
    state = list(range(n))
    munge(state, lengths=lengths, n=n, iterations=1)
    return state[0] * state[1]


def knot_hash(data, n=256):
    state = list(range(n))
    lengths = [ord(x) for x in data] + [17, 31, 73, 47, 23]
    munge(state, lengths, n)
    reduced = [reduce(xor, state[16 * i : 16 * (i + 1)]) for i in range(16)]
    return bytes(reduced).hex()


if __name__ == "__main__":
    print("part a:", part_a(data))
    print("part b:", knot_hash(data))



"""
--- Day 5: Hydrothermal Venture ---
https://adventofcode.com/2021/day/5
"""
from aocd import data
from collections import Counter
from parse import parse

da = Counter()
db = Counter()
for line in data.splitlines():
    x1, y1, x2, y2 = parse("{:d},{:d} -> {:d},{:d}", line).fixed
    xstep = 1 if x1 < x2 else -1
    ystep = 1 if y1 < y2 else -1
    xrange = range(x1, x2 + xstep, xstep)
    yrange = range(y1, y2 + ystep, ystep)
    if x1 == x2:  # vertical line
        for y in yrange:
            da[x1, y] += 1
    elif y1 == y2:  # horizontal line
        for x in xrange:
            da[x, y1] += 1
    elif abs(y2 - y1) == abs(x2 - x1):  # diagonal line
        for x, y in zip(xrange, yrange):
            db[x, y] += 1

print("part a:", sum(v > 1 for v in da.values()))
print("part b:", sum(v > 1 for v in (da + db).values()))



"""
--- Day 9: Stream Processing ---
https://adventofcode.com/2017/day/9
"""
import re

from aocd import data


n_removed = 0
while "!!" in data:
    data = data.replace("!!", "")
data = re.sub("!.", "", data)
n_removed += len(data) - len(re.sub("<.*?>", "", data))
n_removed -= 2 * len(re.findall("<.*?>", data))
data = re.sub("<.*?>", "", data)
score = 0
nesting = 0
for c in data:
    if c == "{":
        nesting += 1
    elif c == "}":
        score += nesting
        nesting -= 1

print("part a:", score)
print("part b:", n_removed)



"""
--- Day 6: Universal Orbit Map ---
https://adventofcode.com/2019/day/6
"""
import networkx as nx
from aocd import data


graph = nx.Graph(x.split(")") for x in data.splitlines())
print("part a:", sum(nx.shortest_path_length(graph, x, "COM") for x in graph))
if "YOU" in graph:
    print("part b:", nx.shortest_path_length(graph, "YOU", "SAN") - 2)



"""
--- Day 3: No Matter How You Slice It ---
https://adventofcode.com/2018/day/3
"""
from types import SimpleNamespace

import numpy as np
from aocd import data
from parse import parse


template = "#{id:d} @ {col:d},{row:d}: {w:d}x{h:d}"
claims = [SimpleNamespace(**parse(template, s).named) for s in data.splitlines()]
W = max(c.col + c.w for c in claims)
H = max(c.row + c.h for c in claims)
A = np.zeros((H, W, len(claims)), dtype=int)

for c in claims:
    A[c.row : c.row + c.h, c.col : c.col + c.w, c.id - 1] = 1
print("part a:", (A.sum(axis=2) > 1).sum())

for c in claims:
    if A[c.row : c.row + c.h, c.col : c.col + c.w, :].sum() == c.w * c.h:
        print("part b:", c.id)
        break



"""
--- Day 16: Ticket Translation ---
https://adventofcode.com/2020/day/16
"""
from aocd import data
from math import prod


def parsed(data):
    fields_raw, my_ticket_raw, nearby_tickets_raw = data.split("\n\n")

    fields = {}
    for field in fields_raw.splitlines():
        name, ranges = field.split(": ")
        fields[name] = []
        for r in ranges.split(" or "):
            start, stop = r.split("-")
            fields[name].append(range(int(start), int(stop) + 1))

    _, my_ticket = my_ticket_raw.splitlines()
    my_ticket = [int(x) for x in my_ticket.split(",")]
    
    _, *nearby_tickets = nearby_tickets_raw.splitlines()
    nearby_tickets = [[int(x) for x in row.split(",")] for row in nearby_tickets]

    return fields, my_ticket, nearby_tickets


def identify_fields(fields, valid_tickets):
    cols = [*zip(*valid_tickets)]  # zip-splat transpose trick
    known = {}
    while len(known) < len(fields):
        for i, col in enumerate(cols):
            possibilities = []
            for name, (r1, r2) in fields.items():
                if name not in known and all(v in r1 or v in r2 for v in col):
                    possibilities.append(name)
            if len(possibilities) == 1:
                [name] = possibilities
                known[name] = i
    return known


def get_valid_tickets(fields, nearby_tickets):
    invalid = set()
    a = 0
    for i, ns in enumerate(nearby_tickets):
        for n in ns:
            for _, ranges in fields.items():
                if any(n in r for r in ranges):
                    break  # valid
            else:
                invalid.add(i)
                a += n
    print("part a:", a)
    valid_tickets = [nn for i, nn in enumerate(nearby_tickets) if i not in invalid]
    return valid_tickets


def translate_my_ticket(fieldname2col, my_ticket):
    return {name: my_ticket[col] for name, col in fieldname2col.items()}


if __name__ == "__main__":
    fields, my_ticket, nearby_tickets = parsed(data)
    valid_tickets = get_valid_tickets(fields, nearby_tickets)
    known = identify_fields(fields, valid_tickets)
    my_ticket_dict = translate_my_ticket(known, my_ticket)
    b = prod(v for k, v in my_ticket_dict.items() if k.startswith("departure"))
    print("part b:", b)



"""
--- Day 22: Crab Combat ---
https://adventofcode.com/2020/day/22
"""
from aocd import data
from collections import deque
from itertools import islice


def play(p1, p2, part="a"):
    seen = set()
    while p1 and p2:
        if part == "b":
            k = *p1, None, *p2
            if k in seen:
                return 1
            seen.add(k)
        c1 = p1.popleft()
        c2 = p2.popleft()
        if part == "b" and c1 <= len(p1) and c2 <= len(p2):
            p1r = deque(islice(p1, 0, c1))
            p2r = deque(islice(p2, 0, c2))
            winner = play(p1r, p2r, part)
        else:
            winner = 1 if c1 > c2 else 2
        if winner == 1:
            p1.append(c1)
            p1.append(c2)
        else:
            p2.append(c2)
            p2.append(c1)
    return winner


def score(hand):
    return sum(i * n for i, n in enumerate(reversed(hand), start=1))


if __name__ == "__main__":
    data1, data2 = data.split("\n\n")
    cards1 = [int(x) for x in data1.splitlines()[1:]]
    cards2 = [int(x) for x in data2.splitlines()[1:]]

    p1a = deque(cards1)
    p2a = deque(cards2)
    play(p1a, p2a)
    print("part a:", score(p1a or p2a))

    p1b = deque(cards1)
    p2b = deque(cards2)
    play(p1b, p2b, part="b")
    print("part b:", score(p1b or p2b))



"""
--- Day 11: Monkey in the Middle ---
https://adventofcode.com/2022/day/11
"""
from aocd import data
from parse import parse
from collections import deque
from heapq import nlargest
from math import prod


template = """\
Monkey {:d}:
  Starting items: {items}
  Operation: new = {x} {op} {y}
  Test: divisible by {div:d}
    If true: throw to monkey {mt:d}
    If false: throw to monkey {mf:d}"""


class Monkey:

    def __init__(self, items, x, op, y, div, mt, mf):
        self.items = items
        self.x = x
        self.op = op
        self.y = y
        self.div = div
        self.mt = mt
        self.mf = mf
        self.inspected = 0

    @classmethod
    def fromchunk(cls, chunk):
        parsed = parse(template, chunk).named
        items = deque(map(int, parsed.pop("items").split(", ")))
        return cls(items, **parsed)

    def inspect(self):
        while self.items:
            i = self.items.popleft()
            x = i if self.x == "old" else int(self.x)
            y = i if self.y == "old" else int(self.y)
            i = x + y if self.op == "+" else x * y
            i = i // 3 if part == "a" else i % denom
            mnext = self.mf if i % self.div else self.mt
            monkeys[mnext].items.append(i)
            self.inspected += 1


for part, nrounds in zip("ab", (20, 10000)):
    monkeys = [Monkey.fromchunk(x) for x in data.split("\n\n")]
    denom = prod([m.div for m in monkeys])
    for round in range(nrounds):
        for monkey in monkeys:
            monkey.inspect()
    print(f"part {part}:", prod(nlargest(2, [m.inspected for m in monkeys])))



"""
--- Day 10: Adapter Array ---
https://adventofcode.com/2020/day/10
"""
from aocd import data
from aoc_wim.stuff import Tribonacci

numbers = [int(x) for x in data.split()]
numbers.append(0)  # outlet
numbers.sort()
Δ = [y - x for x, y in zip(numbers, numbers[1:])]
Δ.append(3)  # device
print("part a:", Δ.count(1) * Δ.count(3))

tri = Tribonacci()
Π = 1
ones = 0
for d in Δ:
    if d == 1:
        ones += 1  # finding longest streak of ones
        continue
    Π *= tri[ones + 2] or 1  # streak broken, cash in
    ones = 0
print("part b:", Π)



"""
--- Day 3: Rucksack Reorganization ---
https://adventofcode.com/2022/day/3
"""
from aocd import data
from wimpy import chunks

a = 0
for line in data.splitlines():
    i = len(line) // 2  # midpoint
    left, right = line[:i], line[i:]
    [c] = set(left).intersection(right)
    a += c.isupper() * 26
    a += ord(c.lower()) - ord("a") + 1

b = 0
for s1, s2, s3 in chunks(data.splitlines()):
    [c] = set(s1).intersection(s2, s3)
    b += c.isupper() * 26
    b += ord(c.lower()) - ord("a") + 1

print("part a:", a)
print("part b:", b)



"""
--- Day 2: 1202 Program Alarm ---
https://adventofcode.com/2019/day/2
"""
from aoc_wim.aoc2019 import IntComputer
from aocd import data


def part_a(data, r1=None, r2=None):
    comp = IntComputer(data)
    if r1 is not None:
        comp.reg[1] = r1
    if r2 is not None:
        comp.reg[2] = r2
    comp.run()
    result = comp.reg[0]
    return result


def part_b(data):
    target = 19690720
    for r1 in range(100):
        for r2 in range(100):
            if part_a(data, r1, r2) == target:
                return 100 * r1 + r2


print(part_a(data, r1=12, r2=2))
print(part_b(data))



"""
--- Day 17: Pyroclastic Flow ---
https://adventofcode.com/2022/day/17
"""
from aocd import data
from aoc_wim.zgrid import ZGrid


sprites = """\
####

.#.
###
.#.

..#
..#
###

#
#
#
#

##
##"""


bricks = []
for sprite in sprites.split("\n\n"):
    g = ZGrid(sprite)
    zs = g.z("#", first=False)
    bricks.append((zs, g.width, g.height))


W = 7
Ta = 2022
Tb = 1000000000000
grid = ZGrid({x + 1j: "#" for x in range(W)})

seen = {}
a = i = top = i_push = 0
while i < Tb:
    i_brick = i % len(bricks)
    if i == Ta:
        a = top
    elif i > Ta:
        # index into bricks, index into moves, current "shape" of top row of grid
        shape = "".join([grid.get(x - top * 1j, ".") for x in range(W)])
        state = i_brick, i_push, shape
        if state in seen:
            i_prev, top_prev = seen[state]
            period = i - i_prev
            growth = top - top_prev
            n = (Tb - i) // period
            new_top = top + n * growth
            new_i = i + n * period
            # copy what the top of the grid looks like, and fast-forward time bigly
            for h in range(20):
                for x in range(W):
                    if x + (h - top) * 1j in grid:
                        grid[x + (h - new_top) * 1j] = "#"
            top = new_top
            i = new_i
            seen.clear()
        seen[state] = i, top

    brick, w, h = bricks[i_brick]
    p0 = 2 - (top + h + 2) * 1j
    g = ZGrid(dict.fromkeys(brick, "#"))
    g.offset(p0)
    x0 = p0.real
    while True:
        push = data[i_push]
        i_push = (i_push + 1) % len(data)
        if push == ">":
            if (x0 + w + 1).real <= W and not any(z + 1 in grid for z in g):
                g.offset(1)
                x0 += 1
        elif push == "<":
            if (x0 - 1).real >= 0 and not any(z - 1 in grid for z in g):
                g.offset(-1)
                x0 -= 1
        if any(z + 1j in grid for z in g):
            grid.update(g)
            top = max(top, int(1 - g.top_left.imag))
            break
        g.offset(1j)
    i += 1


print("part a:", a)
print("part b:", top)



"""
--- Day 6: Signals and Noise ---
https://adventofcode.com/2016/day/6
"""
from collections import Counter
from aocd import data

counters = [Counter(x) for x in zip(*data.splitlines())]
print("part a:", "".join([max(c, key=c.get) for c in counters]))
print("part b:", "".join([min(c, key=c.get) for c in counters]))



"""
--- Day 17: Trick Shot ---
https://adventofcode.com/2021/day/17
"""
from aocd import data
from parse import parse
from aoc_wim.zgrid import ZGrid


def launch(dx, dy, draw=False):
    x = y = y_max = 0
    xys = []
    while y >= y0 and x <= x1:
        x += dx
        y += dy
        xys.append((x, y))
        dx -= 1 if dx > 0 else -1 if dx < 0 else 0  # drag
        dy -= 1  # gravity
        y_max = max(y_max, y)
        if (x, y) in target_area:
            if draw:
                grid = ZGrid({0: "S"} | {complex(*z): "T" for z in target_area})
                grid.draw(flip="y", overlay={complex(*xy): "#" for xy in xys})
            return y_max


x0, x1, y0, y1 = parse("target area: x={:d}..{:d}, y={:d}..{:d}", data).fixed
target_area = {(x, y) for x in range(x0, x1 + 1) for y in range(y0, y1 + 1)}
dx_min = 1
while sum(range(dx_min + 1)) < x0:  # while target area is undershot horizontally
    dx_min += 1   # anything less is too slow to reach the target horizontally
dx_max = x1 + 1   # anything more overshooting horizontally on the first iteration
dy_min = y0 - 1   # anything less undershooting vertically on the first iteration
dy_max = 1 - y0   # anything higher is coming down too hot

a = b = 0
for dx in range(dx_min, dx_max + 1):
    for dy in range(dy_min, dy_max + 1):
        y = launch(dx, dy)
        if y is not None:
            b += 1
            if y > a:
                a = y
                dxy_best = dx, dy

print("part a:", a)
print("part b:", b)



"""
--- Day 25: Cryostasis ---
https://adventofcode.com/2019/day/25
"""
import logging
from collections import deque
import anytree
from aocd import data
from aoc_wim.aoc2019 import IntComputer
from itertools import combinations
from wimpy.util import cached_property


log = logging.getLogger(__name__)


def parse_room(txt):
    txt = txt.strip()
    lines = txt.splitlines()
    if len(lines) < 3:
        log.warning("failed to parse output\n%r", txt)
        raise Exception
    first, description, *rest, last = lines
    if not first.startswith("==") or not last.startswith("Command?"):
        log.warning("failed to parse output\n%r", txt)
        raise Exception
    name = first.strip("= ")
    doors = []
    items = []
    section = None
    for line in rest:
        line = line.strip()
        if not line:
            continue
        if line == "Doors here lead:":
            section = doors
            continue
        if line == "Items here:":
            section = items
            continue
        if line.startswith("- "):
            line = line[2:]
            section.append(line)
            continue
        if not line.startswith("Command?"):
            log.info("suffix: %s", line)
    doors = {}.fromkeys(doors)
    result = {"name": name, "description": description, "doors": doors, "items": items}
    return result


class Room(anytree.NodeMixin):
    def __init__(self, name, description, doors, items):
        self.name = name
        self.description = description
        self.doors = doors
        self.items = items

    def __repr__(self):
        return f"<Room({self.name})>"


class Game:
    def __init__(self):
        self.comp = IntComputer(data)
        self.root = None
        self.comp.input = self
        self.command_queue = deque()
        self.shortcuts = {
            "n": "north",
            "s": "south",
            "e": "east",
            "w": "west",
            "\x1b[A": "north",
            "\x1b[B": "south",
            "\x1b[C": "east",
            "\x1b[D": "west",
        }
        self.inverse = {
            "north": "south",
            "south": "north",
            "east": "west",
            "west": "east",
        }
        self.rooms = {}
        self.current_room = None
        self.prev_direction = None
        self.prev_room = None
        self.inventory = set()
        self.mode = "explore"  # or "unlock"
        self.combo = None

    def shortest_path(self, to):
        to_path = to.path
        path = []
        for room in reversed(self.current_room.path):
            if room in to_path:
                to_path = to_path[to_path.index(room):]
                path.extend(to_path)
                break
            path.append(room)
        return path

    def find_path(self, to=None):
        last_step = None
        if to is None:
            # just go to any unexplored room...
            for room in self.rooms.values():
                if room.name == "Security Checkpoint":
                    continue
                if any(v is None for v in room.doors.values()):
                    to = room
                    last_step = next(k for k, v in room.doors.items() if v is None)
                    break
            else:
                return
        intermediate_rooms = self.shortest_path(to=to)
        path = []
        for left, right in zip(intermediate_rooms, intermediate_rooms[1:]):
            direction = next(d for d, r in left.doors.items() if r is right)
            path.append(direction)
        if last_step is not None:
            path.append(last_step)
        if not path:
            # can happen if you try to find path to room but you're already there
            return
        first_step = path[0]
        return first_step

    @cached_property
    def combos(self):
        data = {}
        inventory = sorted(self.inventory)
        for i in range(len(inventory)):
            for combo in combinations(inventory, i):
                data[combo] = "new"
        return data

    @cached_property
    def unlock_direction(self):
        return next(k for k, v in self.current_room.doors.items() if v is None)

    def pop(self):
        if not self.command_queue:
            # command queue is exhausted, need to update status and find task to do
            command = ""
            output = self.comp.output_as_text

            # entered a room
            if output.strip().startswith("=="):
                parsed_room = parse_room(output)
                here = parsed_room["name"]
                if here == "Pressure-Sensitive Floor":
                    if "ejected" in output:
                        if "Droids on this ship are heavier" in output:
                            self.combos[self.combo] = "too light"
                            self.combo = None
                        elif "Droids on this ship are lighter" in output:
                            self.combos[self.combo] = "too heavy"
                            self.combo = None
                        else:
                            log.warning("shouldn't be here: %r", output)
                log.debug("parsed room data: %s", parsed_room)
                if here not in self.rooms:
                    log.debug("entered new room %r, adding to graph", parsed_room)
                    new_room = self.rooms[here] = Room(**parsed_room)
                    if self.prev_room is not None:
                        new_room.parent = self.prev_room
                        back = self.inverse[self.prev_direction]
                        new_room.doors[back] = self.prev_room
                        self.prev_room.doors[self.prev_direction] = new_room
                    else:
                        self.root = new_room

                self.current_room = self.rooms[here]
                if here == "Pressure-Sensitive Floor":
                    if "ejected" in output:
                        # actually we got kicked out
                        self.current_room = self.rooms["Security Checkpoint"]

            # update inventory of self and rooms
            for line in output.splitlines():
                if line.startswith("You take the "):
                    item = line[13:-1]
                    self.current_room.items.remove(item)
                    self.inventory.add(item)
                if line.startswith("You drop the "):
                    item = line[13:-1]
                    self.inventory.remove(item)
                    self.current_room.items.append(item)

            # echo on
            self.comp.output.clear()
            print(output, end="", flush=True)

            # pick up any crap we find lying around
            if self.mode == "explore":
                for item in self.current_room.items:
                    if item not in blacklist:
                        log.info("autotaking %r", item)
                        command += f"take {item}\n"

            # there was nothing to take, explore around the ship
            if self.mode == "explore" and not command:
                step = self.find_path()
                if step is not None:
                    command = step
                else:
                    # we've explored everything, proceed to checkpoint
                    self.mode = "unlock"
                    command = self.find_path(to=self.rooms["Security Checkpoint"])
                if command is None:
                    command = "inv"
                else:
                    self.prev_direction = command
                    self.prev_room = self.current_room
                log.info("entering command %r", command)

            # continue proceeding to checkpoint
            if self.mode == "unlock":
                if self.current_room.name != "Security Checkpoint":
                    command = self.find_path(to=self.rooms["Security Checkpoint"])
                    self.prev_direction = command
                    self.prev_room = self.current_room

            if self.mode == "unlock" and not command:
                assert self.current_room.name == "Security Checkpoint"
                if self.combo is None:
                    self.combo = next(k for k, v in self.combos.items() if v == "new")
                combo = set(self.combo)
                diff = combo ^ self.inventory
                for item in diff:
                    if item in self.inventory:
                        command = f"drop {item}"
                    else:
                        assert item in combo and item not in self.inventory
                        command = f"take {item}"
                    break
                else:
                    assert combo == self.inventory
                    log.info("trying combo %s", combo)
                    command = self.prev_direction = self.unlock_direction
                    self.prev_room = self.current_room

            # don't know what to do
            if not command:
                # interactive fallback
                command = input()

            if not command:
                # so that pressing "enter" at the prompt adds some useful context
                if self.root:
                    print(anytree.RenderTree(self.root))
                print("current room:", self.current_room.name)
                print(self.current_room.description)
                print("doors:", "/".join(self.current_room.doors))
                print("inventory:", ", ".join(sorted(self.inventory)))

            # a convenience shortcut to take the item here
            if command == "t" and len(self.current_room.items) == 1:
                [item] = self.current_room.items
                command = f"take {item}"

            # allows to use the arrow keys or enter just "n" instead of typing "north"
            command = self.shortcuts.get(command, command)
            if not command.endswith("\n"):
                command += "\n"

            for char in command:
                self.command_queue.appendleft(ord(char))

        return self.command_queue.pop()

    def play(self):
        self.comp.run()
        txt = self.comp.output_as_text
        if "Analysis complete! You may proceed." not in txt:
            log.warning("should be done but: %r", txt)
            raise Exception
        [code] = [word for word in txt.split() if word.isdigit()]
        print("part a", code)


# traps .. don't pick up this crap
blacklist = [
    "photons",
    "infinite loop",
    "molten lava",
    "giant electromagnet",
    "escape pod",
]

game = Game()
game.play()



"""
--- Day 9: All in a Single Night ---
https://adventofcode.com/2015/day/9
"""
from itertools import permutations

from aocd import data


def parsed(data):
    distances = {}
    cities = set()
    for line in data.splitlines():
        a, to, b, equals, n = line.split()
        cities |= {a, b}
        distances[(a, b)] = distances[(b, a)] = int(n)
    return cities, distances


def total_distance(route, distances):
    return sum(distances[p] for p in zip(route[:-1], route[1:]))


def get_routes(data):
    cities, distances = parsed(data)
    return {route: total_distance(route, distances) for route in permutations(cities)}


routes = get_routes(data)
print("part a:", min(routes.values()))
print("part b:", max(routes.values()))
